import pkg_resources

from pycantonese.corpus import hkcancor, read_chat, CHATReader
from pycantonese.jyutping.characters import (
    characters_to_jyutping,
    characters2jyutping,
)
from pycantonese.jyutping.parse_jyutping import parse_jyutping
from pycantonese.jyutping.tipa import jyutping_to_tipa, jyutping2tipa
from pycantonese.jyutping.yale import jyutping_to_yale, jyutping2yale
from pycantonese.pos_tagging.tagger import pos_tag
from pycantonese.stop_words import stop_words
from pycantonese.word_segmentation import segment


__version__ = pkg_resources.get_distribution("pycantonese").version

__all__ = [
    "__version__",
    "CHATReader",
    "characters_to_jyutping",
    "characters2jyutping",
    "hkcancor",
    "jyutping_to_tipa",
    "jyutping_to_yale",
    "jyutping2tipa",
    "jyutping2yale",
    "parse_jyutping",
    "pos_tag",
    "read_chat",
    "stop_words",
    "segment",
]



import json
import os


_THIS_DIR = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(_THIS_DIR, "chars_to_jyutping.json"), encoding="utf8") as f:
    CHARS_TO_JYUTPING = json.load(f)

with open(os.path.join(_THIS_DIR, "lettered.json"), encoding="utf8") as f:
    LETTERED = json.load(f)



import unittest

import pytest
from pylangacq.tests.test_chat import BaseTestCHATReader

import pycantonese


skip_because_tokens_are_different = pytest.mark.skip(
    reason=(
        "PyCantonese's Token class has the extra ``jyutping`` attribute and "
        "turns meaningless ``mor`` into ``None`` (not done at ``pylangacq.Reader``)."
    ),
)


class TestPyCantoneseCHATReader(BaseTestCHATReader, unittest.TestCase):
    """Test the CHAT reader methods by using ``pycantonese.CHATReader``."""

    reader_class = pycantonese.CHATReader

    @skip_because_tokens_are_different
    def test_ipsyn(self):
        pass

    @skip_because_tokens_are_different
    def test_tokens(self):
        pass

    @skip_because_tokens_are_different
    def test_tokens_by_utterances(self):
        pass

    @skip_because_tokens_are_different
    def test_utterances(self):
        pass



import unicodedata

from pycantonese.jyutping.parse_jyutping import parse_jyutping
from pycantonese.util import _deprecate


ONSETS_YALE = {
    "b": "b",
    "d": "d",
    "g": "g",
    "gw": "gw",
    "z": "j",
    "p": "p",
    "t": "t",
    "k": "k",
    "kw": "k",
    "c": "ch",
    "m": "m",
    "n": "n",
    "ng": "ng",
    "f": "f",
    "h": "h",
    "s": "s",
    "l": "l",
    "w": "w",
    "j": "y",
    "": "",
}

NUCLEI_YALE = {
    "aa": "aa",
    "a": "a",
    "i": "i",
    "yu": "yu",
    "u": "u",
    "oe": "eu",
    "e": "e",
    "eo": "eu",
    "o": "o",
    "m": "m",
    "ng": "ng",
}

CODAS_YALE = {
    "p": "p",
    "t": "t",
    "k": "k",
    "m": "m",
    "n": "n",
    "ng": "ng",
    "i": "i",
    "u": "u",
    "": "",
}


def jyutping_to_yale(jp_str, as_list=True):
    """Convert Jyutping romanization into Yale romanization.

    .. versionadded:: 3.0.0
        This function replaces the deprecated equivalent ``jyutping2yale``.

    .. versionchanged:: 3.0.0
        ``as_list`` has its default value switched from ``False`` to ``True``,
        so that by default the function returns a list, which is in line with
        the other "jyutping_to_X" functions.

    Parameters
    ----------
    jp_str : str
        Jyutping romanization for one or multiple characters
    as_list : bool, optional
        If False (default is True), the output is a string with a single quote
        ``'`` to disambiguate unclear syllable boundaries (e.g., a consonant
        or the low-tone marker "h" being ambiguous as an onset or as
        part of the previous syllable).

    Returns
    -------
    list[str], or str if as_list is False

    Raises
    ------
    ValueError
        If the Jyutping romanization is illegal (e.g., with unrecognized
        elements).

    Examples
    --------
    >>> jyutping_to_yale("gwong2dung1waa2")  # 廣東話, Cantonese
    ['gwóng', 'dūng', 'wá']
    >>> jyutping_to_yale("gwong2dung1waa2", as_list=False)
    'gwóngdūngwá'
    >>>
    >>> # 'heihauh' would be ambiguous between hei3hau6 and hei6au6.
    >>> jyutping_to_yale("hei3hau6", as_list=False)  # 氣候, climate
    "hei'hauh"
    """
    jp_parsed_list = parse_jyutping(jp_str)
    yale_list = []

    for jp_parsed in jp_parsed_list:
        onset = ONSETS_YALE[jp_parsed.onset]
        nucleus = NUCLEI_YALE[jp_parsed.nucleus]
        coda = CODAS_YALE[jp_parsed.coda]
        tone = jp_parsed.tone  # still in parse_jyutping

        # jyutping2yale system uses "h" to mark the three low tones
        if tone in {"4", "5", "6"}:
            low_tone_h = "h"
        else:
            low_tone_h = ""

        # in jyutping2yale, long "aa" vowel with no coda is denoted by "a"
        if nucleus == "aa" and coda == "":
            nucleus = "a"

        # when nucleus is "yu"...
        # 1. disallow "yyu" (when onset is "y")
        # 2. change nucleus "yu" into "u" -- this is a hack for adding tone
        #       diacritic, since we don't want "y" to bear the diacritic
        if nucleus == "yu":
            if onset == "y":
                onset = ""
            nucleus = "u"

        # when nucleus is "ng"
        # the tone diacritic has to be on "g" but not "n"
        # now we pretend that the nucleus is "g", and will prepend the "n" back
        # at the end
        if nucleus == "ng":
            nucleus = "g"

        # add the jyutping2yale tone diacritic to the first nucleus letter
        # parse_jyutping tone 1      --> add macron
        # parse_jyutping tone 2 or 5 --> add acute
        # parse_jyutping tone 4      --> add grave
        # parse_jyutping tone 3 or 6 --> (no diacritic)
        # If the accented letter doesn't exist in unicode, use the combining
        # accent instead.

        letter = nucleus[0]  # nucleus 1st letter
        unicode_letter_name = unicodedata.name(letter)
        if tone == "1":
            try:
                letter_with_diacritic = unicodedata.lookup(
                    unicode_letter_name + " WITH MACRON"
                )
            except KeyError:
                letter_with_diacritic = letter + "\u0304"
        elif tone in {"2", "5"}:
            try:
                letter_with_diacritic = unicodedata.lookup(
                    unicode_letter_name + " WITH ACUTE"
                )
            except KeyError:
                letter_with_diacritic = letter + "\u0301"
        elif tone == "4":
            try:
                letter_with_diacritic = unicodedata.lookup(
                    unicode_letter_name + " WITH GRAVE"
                )
            except KeyError:
                letter_with_diacritic = letter + "\u0300"
        else:
            # either tone 3 or tone 6
            letter_with_diacritic = letter
        nucleus = letter_with_diacritic + nucleus[1:]

        # add back "y" if the nucleus is "yu"
        # ("y" was taken away for convenience in adding tone diacritic)
        if jp_parsed.nucleus == "yu":
            nucleus = "y" + nucleus

        # add back "n" if the nucleus is "ng"
        # ('n' was taken away so that tone diacritic is on "g" but not "n")
        if jp_parsed.nucleus == "ng":
            nucleus = "n" + nucleus

        # parse_jyutping final "eu" should be jyutping2yale "ew" (not "eu")
        if coda == "u" and nucleus == "e":
            coda = "w"

        # save the resultant jyutping2yale
        if coda in {"i", "u", "w"} and tone in {"4", "5", "6"}:
            yale = onset + nucleus + coda + low_tone_h
        else:
            yale = onset + nucleus + low_tone_h + coda
        yale_list.append(yale)

    if as_list:
        return yale_list

    # Output yale_list as a string
    # Check if there's potential ambiguity when Yale strings are concatenated

    # Ambiguity case 1:
    #   1st syllable coda is one of the "ambiguous_consonants"
    #   and 2nd syllable starts with a vowel *letter*

    # Ambiguity case 2:
    #   1st syllable has no coda and 2nd syllable starts with one of the
    #   "ambiguous_consonants"
    #   e.g., hei3hau6 'climate' --> heihauh
    #   (middle "h" for tone in 1st syllable or being onset of 2nd syllable?)

    if len(yale_list) == 0:
        return ""
    elif len(yale_list) == 1:
        return yale_list[0]

    ambiguous_consonants = {"h", "p", "t", "k", "m", "n", "ng"}
    vowel_letters = {
        "a",
        "e",
        "i",
        "o",
        "u",
        "á",
        "é",
        "í",
        "ó",
        "ú",
        "à",
        "è",
        "ì",
        "ò",
        "ù",
        "ā",
        "ē",
        "ī",
        "ō",
        "ū",
    }

    output_str = ""

    for i in range(len(yale_list) - 1):
        yale1 = yale_list[i]
        yale2 = yale_list[i + 1]

        ambiguous = False

        # test case 1:
        if _endswithoneof(yale1, ambiguous_consonants) and _startswithoneof(
            yale2, vowel_letters
        ):
            ambiguous = True

        # test case 2:
        if (
            not ambiguous
            and not _endswithoneof(yale1, ambiguous_consonants)
            and _startswithoneof(yale2, ambiguous_consonants)
        ):
            ambiguous = True

        output_str += yale1

        if ambiguous:
            output_str += "'"

    output_str += yale_list[-1]

    return output_str


@_deprecate("jyutping2yale", "jyutping_to_yale", "3.0.0", "4.0.0")
def jyutping2yale(*args, **kwargs):
    """Same as jyutping_to_yale.

    .. deprecated:: 3.0.0
    """
    return jyutping_to_yale(*args, **kwargs)


def _startswithoneof(inputstr, seq):
    """
    Check if *inputstr* starts with one of the items in seq. If it does, return
        the item that it starts with. If it doe not, return ``None``.

    :param inputstr: input string

    :param seq: sequences of items to check

    :return: the item the the input string starts with (``None`` if not found)

    :rtype: str or None
    """
    seq = set(seq)
    for item in seq:
        if inputstr.startswith(item):
            return item
    else:
        return None


def _endswithoneof(inputstr, seq):
    """
    Check if *inputstr* ends with one of the items in seq. If it does, return
        the item that it ends with. If it doe not, return ``None``.

    :param inputstr: input string

    :param seq: sequences of items to check

    :return: the item the the input string ends with (``None`` if not found)

    :rtype: str or None
    """
    seq = set(seq)
    for item in seq:
        if inputstr.endswith(item):
            return item
    else:
        return None



import os

from pycantonese.tests import REPO_DIR


def test_changelog_has_unreleased_section():
    """The 'Unreleased' section is needed to make the doc script work."""
    with open(os.path.join(REPO_DIR, "CHANGELOG.md"), encoding="utf8") as f:
        changelog = f.read()
    assert "\n## [Unreleased]\n" in changelog



import os
import re

import pycantonese
from pycantonese.tests import REPO_DIR


def test_version_number_match_with_changelog():
    """__version__ and CHANGELOG.md match for the latest version number."""
    changelog = open(os.path.join(REPO_DIR, "CHANGELOG.md"), encoding="utf-8").read()
    # latest version number in changelog = the 1st occurrence of '[x.y.z]'
    version_in_changelog = (
        re.search(r"\[\d+\.\d+\.\d+\]", changelog).group().strip("[]")
    )
    if "dev" in pycantonese.__version__:
        # CHANGELOG.md doesn't update the section headings for dev versions,
        # and so pycantonese.__version__ with "dev" (e.g., "3.1.0dev1")
        # wouldn't match any version headings in CHANGELOG.md.
        return
    assert pycantonese.__version__ == version_in_changelog, (
        "Make sure both __version__ and CHANGELOG are updated to match the "
        "latest version number"
    )



import pytest

from pycantonese import jyutping_to_yale
from pycantonese.jyutping.parse_jyutping import ONSETS, NUCLEI, CODAS
from pycantonese.jyutping.yale import (
    ONSETS_YALE,
    NUCLEI_YALE,
    CODAS_YALE,
)


def test_correct_onset_set():
    assert set(ONSETS_YALE.keys()) == ONSETS


def test_correct_nucleus_set():
    assert set(NUCLEI_YALE.keys()) == NUCLEI


def test_correct_coda_set():
    assert set(CODAS_YALE.keys()) == CODAS


@pytest.mark.parametrize("input_", ["", None])
def test_null_input(input_):
    assert jyutping_to_yale(input_) == []


@pytest.mark.parametrize("input_", ["", None])
def test_null_input_as_list_false(input_):
    assert jyutping_to_yale(input_, as_list=False) == ""


def test_jyutping_to_yale_m4goi1():
    assert jyutping_to_yale("m4goi1") == ["m̀h", "gōi"]


def test_jyutping_to_yale_gwong2dung1waa2():
    assert jyutping_to_yale("gwong2dung1waa2") == ["gwóng", "dūng", "wá"]



from collections import Counter, defaultdict
from functools import lru_cache

from pycantonese.corpus import hkcancor, Token
from pycantonese.data.rime_cantonese import CHARS_TO_JYUTPING, LETTERED
from pycantonese.jyutping.parse_jyutping import parse_jyutping
from pycantonese.word_segmentation import segment
from pycantonese.util import _split_chars_with_alphanum, _deprecate


@lru_cache(maxsize=1)
def _get_words_characters_to_jyutping():
    corpus = hkcancor()
    words_to_jyutping_counters = defaultdict(Counter)
    characters_to_jyutping_counters = defaultdict(Counter)

    for token in corpus.tokens(by_utterances=False):
        token: Token
        word = token.word
        jyutping = token.jyutping

        if not jyutping or not word:
            continue
        try:
            parsed_jp = parse_jyutping(jyutping)
        except ValueError:
            continue
        if len(word) != len(parsed_jp):
            continue
        words_to_jyutping_counters[word][jyutping] += 1
        for char, jp in zip(word, parsed_jp):
            characters_to_jyutping_counters[char][str(jp)] += 1

    words_to_jyutping = {}
    for word, jyutping_counter in words_to_jyutping_counters.items():
        jp = jyutping_counter.most_common(1)[0][0]
        words_to_jyutping[word] = jp
    characters_to_jyutping = {}
    for character, jyutping_counter in characters_to_jyutping_counters.items():
        jp = jyutping_counter.most_common(1)[0][0]
        characters_to_jyutping[character] = jp

    words_to_jyutping = {
        # The ordering of the following dicts matters. The rime-cantonese
        # data may contain what's been re-segmented by this repo, and may
        # contain jyutping pronunciations for particular characters that
        # are only used in those contexts. The data from HKCanCor should comes
        # last to act as the default to override such cases.
        **{k: v for k, v in LETTERED.items() if len(_split_chars_with_alphanum(k)) > 1},
        **{k: v for k, v in CHARS_TO_JYUTPING.items() if len(k) > 1},
        **words_to_jyutping,
    }

    # TODO: Extract characters from CHARS_TO_JYUTPING and LETTERED
    #    and add them to characters_to_jyutping
    characters_to_jyutping = {
        # The ordering of the following dicts matters. The rime-cantonese
        # data may contain what's been re-segmented by this repo, and may
        # contain jyutping pronunciations for particular characters that
        # are only used in those contexts. The data from HKCanCor should comes
        # last to act as the default to override such cases.
        **{k: v for k, v in LETTERED.items() if len(k) == 1},
        **{k: v for k, v in CHARS_TO_JYUTPING.items() if len(k) == 1},
        **characters_to_jyutping,
    }

    return words_to_jyutping, characters_to_jyutping


def characters_to_jyutping(chars):
    """Convert Cantonese characters into Jyutping romanization.

    The conversion model is based on the HKCanCor corpus and rime-cantonese
    data. Any unseen Cantonese character (or punctuation mark,
    for that matter) is represented by `None` in the output.

    The output is a list of segmented words, where each word is a 2-tuple of
    (Cantonese characters, Jyutping romanization).

    .. versionadded:: 3.0.0
        This function replaces the deprecated equivalent
        ``characters2jyutping``.

    .. versionchanged:: 3.0.0
        The returned valued is now a list of segmented words,
        where each word is a 2-tuple of (Cantonese characters, Jyutping).
        Previously, it was a list of Jyutping strings for the individual
        Cantonese characters.

    Parameters
    ----------
    chars : str
        A string of Cantonese characters.

    Returns
    -------
    list[tuple[str]]

    Examples
    --------
    >>> characters_to_jyutping("香港人講廣東話。")  # Hongkongers speak Cantonese.
    [('香港人', 'hoeng1gong2jan4'), ('講', 'gong2'), ('廣東話', 'gwong2dung1waa2'), ('。', None)]
    """  # noqa: E501
    if not chars:
        return []
    words_to_jyutping, chars_to_jyutping = _get_words_characters_to_jyutping()
    result = []
    for word in segment(chars):
        try:
            jp = words_to_jyutping[word]
        except KeyError:
            jp = ""
            for char in word:
                try:
                    jp += chars_to_jyutping[char]
                except KeyError:
                    jp = None
                    break
        result.append((word, jp))
    return result


@_deprecate("characters2jyutping", "characters_to_jyutping", "3.0.0", "4.0.0")
def characters2jyutping(*args, **kwargs):
    """Same as characters_to_jyutping.

    .. deprecated:: 3.0.0
    """
    return characters_to_jyutping(*args, **kwargs)



from pycantonese.corpus import Token
from pycantonese.tests.test_corpus import HKCANCOR


def test_find_verbs_in_hkcancor():
    all_verbs = HKCANCOR.search(pos="^V")
    assert len(all_verbs) == 29726
    assert all_verbs[:10] == [
        Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
        Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
        Token(word="旅行", pos="VN", jyutping="leoi5hang4", mor=None, gra=None),
        Token(word="有冇", pos="V1", jyutping="jau5mou5", mor=None, gra=None),
        Token(word="要", pos="VU", jyutping="jiu3", mor=None, gra=None),
        Token(word="有得", pos="VU", jyutping="jau5dak1", mor=None, gra=None),
        Token(word="冇得", pos="VU", jyutping="mou5dak1", mor=None, gra=None),
        Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
        Token(word="係", pos="V", jyutping="hai6", mor=None, gra=None),
        Token(word="係", pos="V", jyutping="hai6", mor=None, gra=None),
    ]


def test_by_tokens_false():
    all_verbs = HKCANCOR.search(pos="^V", by_tokens=False)
    assert len(all_verbs) == 29726
    expected = ["去", "去", "旅行", "有冇", "要", "有得", "冇得", "去", "係", "係"]
    assert all_verbs[:10] == expected


def test_by_utterances_true():
    all_verbs = HKCANCOR.search(pos="^V", by_utterances=True)
    assert len(all_verbs) == 29726
    print(all_verbs[:2])
    assert all_verbs[:2] == [
        [
            Token(word="喂", pos="E", jyutping="wai3", mor=None, gra=None),
            Token(word="遲", pos="A", jyutping="ci4", mor=None, gra=None),
            Token(word="啲", pos="U", jyutping="di1", mor=None, gra=None),
            Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
            Token(word="唔", pos="D", jyutping="m4", mor=None, gra=None),
            Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
            Token(word="旅行", pos="VN", jyutping="leoi5hang4", mor=None, gra=None),
            Token(word="啊", pos="Y", jyutping="aa3", mor=None, gra=None),
            Token(word="?", pos="?", jyutping=None, mor=None, gra=None),
        ],
        [
            Token(word="喂", pos="E", jyutping="wai3", mor=None, gra=None),
            Token(word="遲", pos="A", jyutping="ci4", mor=None, gra=None),
            Token(word="啲", pos="U", jyutping="di1", mor=None, gra=None),
            Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
            Token(word="唔", pos="D", jyutping="m4", mor=None, gra=None),
            Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
            Token(word="旅行", pos="VN", jyutping="leoi5hang4", mor=None, gra=None),
            Token(word="啊", pos="Y", jyutping="aa3", mor=None, gra=None),
            Token(word="?", pos="?", jyutping=None, mor=None, gra=None),
        ],
    ]


def test_word_range():
    all_verbs = HKCANCOR.search(pos="^V", word_range=(1, 2))
    assert len(all_verbs) == 29726
    assert all_verbs[:2] == [
        [
            Token(word="啲", pos="U", jyutping="di1", mor=None, gra=None),
            Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
            Token(word="唔", pos="D", jyutping="m4", mor=None, gra=None),
            Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
        ],
        [
            Token(word="唔", pos="D", jyutping="m4", mor=None, gra=None),
            Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
            Token(word="旅行", pos="VN", jyutping="leoi5hang4", mor=None, gra=None),
            Token(word="啊", pos="Y", jyutping="aa3", mor=None, gra=None),
        ],
    ]


def test_utterance_range():
    all_verbs = HKCANCOR.search(pos="^V", utterance_range=(0, 1))
    assert len(all_verbs) == 29726
    assert all_verbs[:2] == [
        [
            [
                Token(word="喂", pos="E", jyutping="wai3", mor=None, gra=None),
                Token(word="遲", pos="A", jyutping="ci4", mor=None, gra=None),
                Token(word="啲", pos="U", jyutping="di1", mor=None, gra=None),
                Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
                Token(word="唔", pos="D", jyutping="m4", mor=None, gra=None),
                Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
                Token(word="旅行", pos="VN", jyutping="leoi5hang4", mor=None, gra=None),
                Token(word="啊", pos="Y", jyutping="aa3", mor=None, gra=None),
                Token(word="?", pos="?", jyutping=None, mor=None, gra=None),
            ],
            [
                Token(word="你", pos="R", jyutping="nei5", mor=None, gra=None),
                Token(word="老公", pos="N", jyutping="lou5gung1", mor=None, gra=None),
                Token(word="有冇", pos="V1", jyutping="jau5mou5", mor=None, gra=None),
                Token(word="平", pos="A", jyutping="peng4", mor=None, gra=None),
                Token(word="機票", pos="N", jyutping="gei1piu3", mor=None, gra=None),
                Token(word="啊", pos="Y", jyutping="aa3", mor=None, gra=None),
                Token(word="?", pos="?", jyutping=None, mor=None, gra=None),
            ],
        ],
        [
            [
                Token(word="喂", pos="E", jyutping="wai3", mor=None, gra=None),
                Token(word="遲", pos="A", jyutping="ci4", mor=None, gra=None),
                Token(word="啲", pos="U", jyutping="di1", mor=None, gra=None),
                Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
                Token(word="唔", pos="D", jyutping="m4", mor=None, gra=None),
                Token(word="去", pos="V", jyutping="heoi3", mor=None, gra=None),
                Token(word="旅行", pos="VN", jyutping="leoi5hang4", mor=None, gra=None),
                Token(word="啊", pos="Y", jyutping="aa3", mor=None, gra=None),
                Token(word="?", pos="?", jyutping=None, mor=None, gra=None),
            ],
            [
                Token(word="你", pos="R", jyutping="nei5", mor=None, gra=None),
                Token(word="老公", pos="N", jyutping="lou5gung1", mor=None, gra=None),
                Token(word="有冇", pos="V1", jyutping="jau5mou5", mor=None, gra=None),
                Token(word="平", pos="A", jyutping="peng4", mor=None, gra=None),
                Token(word="機票", pos="N", jyutping="gei1piu3", mor=None, gra=None),
                Token(word="啊", pos="Y", jyutping="aa3", mor=None, gra=None),
                Token(word="?", pos="?", jyutping=None, mor=None, gra=None),
            ],
        ],
    ]



"""Test Jyutping parsing.

Some tests mirror what the documentation demonstrates.
If these tests fail, the documentation should probably be updated as well
(and the bugs should be fixed, if any).
"""


import pytest

from pycantonese import parse_jyutping
from pycantonese.jyutping.parse_jyutping import Jyutping


def test_basic_case_gwong2dung1waa2():
    assert parse_jyutping("gwong2dung1waa2") == [
        Jyutping("gw", "o", "ng", "2"),
        Jyutping("d", "u", "ng", "1"),
        Jyutping("w", "aa", "", "2"),
    ]


def test_wrong_data_type():
    with pytest.raises(ValueError):
        parse_jyutping(123)


@pytest.mark.parametrize("input_", ["", None])
def test_null_input(input_):
    assert parse_jyutping(input_) == []


def test_syllabic_nasals():
    # TODO assert parse_jyutping('hm4') == [('h', 'm', '', '4')]
    assert parse_jyutping("ng5") == [Jyutping("", "ng", "", "5")]
    assert parse_jyutping("m4") == [Jyutping("", "m", "", "4")]
    assert parse_jyutping("n3") == [Jyutping("", "n", "", "3")]


def test_invalid_tone():
    with pytest.raises(ValueError) as e:
        parse_jyutping("lei7")
    assert "tone error" in str(e.value)


def test_no_tone():
    with pytest.raises(ValueError) as e:
        parse_jyutping("lei")
    assert "tone error" in str(e.value)


def test_fewer_than_2_characters():
    with pytest.raises(ValueError) as e:
        parse_jyutping("3")
    assert "fewer than 2 characters" in str(e.value)


def test_invalid_coda():
    with pytest.raises(ValueError) as e:
        parse_jyutping("leil3")
    assert "coda error" in str(e.value)


def test_invalid_nucleus():
    with pytest.raises(ValueError) as e:
        parse_jyutping("sk3")
    assert "nucleus error" in str(e.value)


def test_invalid_onset():
    with pytest.raises(ValueError) as e:
        parse_jyutping("shaa1")
    assert "onset error" in str(e.value)


def test_coda_ng():
    assert parse_jyutping("hoeng1") == [Jyutping("h", "oe", "ng", "1")]


def test_no_noda():
    assert parse_jyutping("gaa1") == [Jyutping("g", "aa", "", "1")]



from pycantonese import stop_words


_DEFAULT_STOP_WORDS = stop_words()


def test_stop_words():
    _stop_words = stop_words()
    assert "唔" in _stop_words


def test_stop_words_add_one_word():
    _stop_words = stop_words(add="foobar")
    assert "foobar" in _stop_words
    assert len(_stop_words) - len(_DEFAULT_STOP_WORDS) == 1


def test_stop_words_remove_one_word():
    _stop_words = stop_words(remove="唔")
    assert "唔" not in _stop_words
    assert len(_DEFAULT_STOP_WORDS) - len(_stop_words) == 1


def test_stop_words_add_multiple_words():
    _stop_words = stop_words(add=["foo", "bar", "baz"])
    assert {"foo", "bar", "baz"}.issubset(_stop_words)
    assert len(_stop_words) - len(_DEFAULT_STOP_WORDS) == 3


def test_stop_words_remove_multiple_words():
    _stop_words = stop_words(remove=["唔", "乜嘢", "其他"])
    assert not {"唔", "乜嘢", "其他"}.issubset(_stop_words)
    assert len(_DEFAULT_STOP_WORDS) - len(_stop_words) == 3



"""Chinese full-length punctuation marks.

See: https://en.wikipedia.org/wiki/Chinese_punctuation
"""

_PUNCTUATION_MARKS = """
，
。
！
？
；
：
（
）
「
」
［
］
【
】
《
》
〈
〉
、
‧
…
—
～
"""

_PUNCTUATION_MARKS = frozenset(_PUNCTUATION_MARKS.strip())



import collections
import functools
import logging
import os
import pickle
import random

from typing import Dict

from pycantonese.pos_tagging.punctuation_marks import _PUNCTUATION_MARKS
from pycantonese.pos_tagging.hkcancor_to_ud import hkcancor_to_ud


# Use the highest pickle protocol version that's compatible for all supported
# Python versions.
# Protocol version 4 was added in Python 3.4.
# Protocol version 5 was added in Python 3.8.
# Reference: https://docs.python.org/3/library/pickle.html#data-stream-format
_PICKLE_PROTOCOL = 4

_THIS_DIR = os.path.dirname(os.path.abspath(__file__))
_PICKLE_PATH = os.path.join(_THIS_DIR, "tagger.pickle")


class _AveragedPerceptron:
    """An averaged perceptron.

    This is a modified version based on the textblob-aptagger codebase
    (MIT license), with original implementation by Matthew Honnibal:
    https://github.com/sloria/textblob-aptagger/blob/266fa1c22daaff7c60577efa8577f1b6ce2f7f70/textblob_aptagger/_perceptron.py
    """

    def __init__(self):
        # Each feature (key) gets its own weight vector (value).
        self.weights: Dict[str, Dict[str, float]] = {}
        self.classes = set()
        # The accumulated values, for the averaging. These will be keyed by
        # feature/class tuples
        self._totals = collections.defaultdict(int)
        # The last time the feature was changed, for the averaging. Also
        # keyed by feature/class tuples
        # (tstamps is short for timestamps)
        self._tstamps = collections.defaultdict(int)
        # Number of instances seen
        self.i = 0

    def predict(self, features):
        """Return the best label for the given features.

        It's computed based on the dot-product between the features and
        current weights.
        """
        scores = collections.defaultdict(float)
        for feat, value in features.items():
            if feat not in self.weights or value == 0:
                continue
            weights = self.weights[feat]
            for label, weight in weights.items():
                scores[label] += value * weight
        # Do a secondary alphabetic sort, for stability
        return max(self.classes, key=lambda label: (scores[label], label))

    def update(self, truth, guess, features):
        """Update the feature weights."""

        def upd_feat(c, f, w, v):
            param = (f, c)
            self._totals[param] += (self.i - self._tstamps[param]) * w
            self._tstamps[param] = self.i
            self.weights[f][c] = w + v

        self.i += 1
        if truth == guess:
            return None
        for f in features:
            weights = self.weights.setdefault(f, {})
            upd_feat(truth, f, weights.get(truth, 0.0), 1.0)
            upd_feat(guess, f, weights.get(guess, 0.0), -1.0)

    def average_weights(self):
        """Average weights from all iterations."""
        for feat, weights in self.weights.items():
            new_feat_weights = {}
            for clas, weight in weights.items():
                param = (feat, clas)
                total = self._totals[param]
                total += (self.i - self._tstamps[param]) * weight
                averaged = round(total / float(self.i), 3)
                if averaged:
                    new_feat_weights[clas] = averaged
            self.weights[feat] = new_feat_weights


class POSTagger:
    """A part-of-speech tagger.

    This is a modified version based on the textblob-aptagger codebase
    (MIT license), with original implementation by Matthew Honnibal:
    https://github.com/sloria/textblob-aptagger/blob/266fa1c22daaff7c60577efa8577f1b6ce2f7f70/textblob_aptagger/taggers.py
    """

    START = ["-START-", "-START2-"]
    END = ["-END-", "-END2-"]

    def __init__(self, *, frequency_threshold=10, ambiguity_threshold=0.95, n_iter=5):
        """Initialize a part-of-speech tagger.

        Parameters
        ----------
        frequency_threshold : int, optional
            A good number of words are almost unambiguously associated with
            a given tag. If these words have a frequency of occurrence above
            this threshold in the training data, they are directly associated
            with their tag in the model.
        ambiguity_threshold : float, optional
            A good number of words are almost unambiguously associated with
            a given tag. If the ratio of (# of occurrences of this word with
            this tag) / (# of occurrences of this word) in the training data
            is equal to or greater than this threshold, then this word is
            directly associated with the tag in the model.
        n_iter : int, optional
            Number of times the training phase iterates through the data.
            At each new iteration, the data is randomly shuffled.
        """
        self.frequency_threshold = frequency_threshold
        self.ambiguity_threshold = ambiguity_threshold
        self.n_iter = n_iter

        self.model = _AveragedPerceptron()
        self.tagdict = {}
        self.classes = set()

        # HKCanCor doesn't have the Chinese full-width punctuation marks.
        self.tagdict.update({punct: punct for punct in _PUNCTUATION_MARKS})

    def tag(self, words):
        """Tag the words.

        Parameters
        ----------
        words : list[str]
            A segmented sentence or phrase, where each word is Cantonese
            characters.

        Returns
        -------
        list[str]
            The list of predicted tags.
        """
        prev, prev2 = self.START
        tags = []
        context = self.START + words + self.END
        for i, word in enumerate(words):
            tag = self.tagdict.get(word)
            if not tag:
                features = self._get_features(i, word, context, prev, prev2)
                tag = self.model.predict(features)
            tags.append(tag)
            prev2 = prev
            prev = tag
        return tags

    def train(self, tagged_sents, save=None):
        """Train a model.

        Parameters
        ----------
        tagged_sents : list[list[tuple[str, str]]]
            A list of segmented and tagged sentences for training.
        save : str, optional
            If given, save the trained model as a pickle at this path.
        """
        self._make_tagdict(tagged_sents)
        self.model.classes = self.classes

        prev, prev2 = self.START
        for iter_ in range(self.n_iter):
            c = 0
            n = 0
            for tagged_sent in tagged_sents:
                context = self.START + [w for w, _ in tagged_sent] + self.END
                for i, (word, tag) in enumerate(tagged_sent):
                    try:
                        guess = self.tagdict[word]
                    except KeyError:
                        feats = self._get_features(i, word, context, prev, prev2)
                        guess = self.model.predict(feats)
                        self.model.update(tag, guess, feats)
                    prev2 = prev
                    prev = guess
                    c += guess == tag
                    n += 1
            random.shuffle(tagged_sents)
            logging.info("Iter %d: %d / %d = %f", iter_, c, n, c / n)
        self.model.average_weights()

        if save is not None:
            pickle.dump(
                (self.model.weights, self.tagdict, self.classes),
                open(save, "wb"),
                protocol=_PICKLE_PROTOCOL,
            )

    def load(self, path):
        """Load a pickled model.

        Parameters
        ----------
        path : str
            The path where the pickled model is located.
        """
        try:
            w_td_c = pickle.load(open(path, "rb"))
        except IOError:
            raise FileNotFoundError(f"Can't locate tagger model {path}")
        except:  # noqa
            raise EnvironmentError(
                f"A file is detected at {path}, but it cannot be read as a "
                "a tagger model. "
                "Either the tagger model file object is corrupted for some reason, "
                "or - perhaps more likely - you're running pycantonese from a local "
                "git repo (e.g., when you are doing dev work) and that you do not have "
                "Git LFS installed on your system. "
                "In the latter case, please install Git LFS "
                "(https://git-lfs.github.com/) and re-install pycantonese."
            )
        self.model.weights, self.tagdict, self.classes = w_td_c
        self.model.classes = self.classes

    def _get_features(self, i, word, context, prev, prev2):
        """Map tokens into a feature representation, implemented as a
        {hashable: float} dict. If the features change, a new model must be
        trained.
        """

        def add(name, *args):
            features[" ".join((name,) + tuple(args))] += 1

        i += len(self.START)
        features = collections.defaultdict(int)

        # It's useful to have a constant feature,
        # which acts sort of like a prior.
        add("bias")

        add("i word's first char", word[0])
        add("i word's final char", word[-1])

        add("i-1 word's first char", context[i - 1][0])
        add("i-1 word's final char", context[i - 1][-1])
        add("i-1 tag", prev)

        add("i-2 word's first char", context[i - 2][0])
        add("i-2 word's final char", context[i - 2][-1])
        add("i-2 tag", prev2)

        add("i+1 word's first char", context[i + 1][0])
        add("i+1 word's final char", context[i + 1][-1])

        add("i+2 word's first char", context[i - 2][0])
        add("i+2 word's final char", context[i - 2][-1])

        return features

    def _make_tagdict(self, tagged_sents):
        """Make a tag dictionary for single-tag words."""
        counts = collections.defaultdict(lambda: collections.defaultdict(int))
        for tagged_sent in tagged_sents:
            for word, tag in tagged_sent:
                counts[word][tag] += 1
                self.classes.add(tag)
        words = set()
        for word, tag_freqs in counts.items():
            words.add(word)
            tag, mode = max(tag_freqs.items(), key=lambda item: item[1])
            n = sum(tag_freqs.values())
            above_freq_threshold = n >= self.frequency_threshold
            unambiguous = (mode / n) >= self.ambiguity_threshold
            if above_freq_threshold and unambiguous:
                self.tagdict[word] = tag
        logging.info("%d unique words in the training data", len(words))
        logging.info("%d tags in this tagset", len(self.classes))
        logging.info("%d words are treated as having a unique tag", len(self.tagdict))


@functools.lru_cache(maxsize=1)
def _get_tagger():
    tagger = POSTagger()
    tagger.load(_PICKLE_PATH)
    return tagger


def pos_tag(words, tagset="universal"):
    """Tag the words for their parts of speech.

    The part-of-speech tagger uses an averaged perceptron model,
    and is trained by the HKCanCor data.

    .. versionadded:: 3.1.0

    Parameters
    ----------
    words : list[str]
        A segmented sentence or phrase, where each word is a string of
        Cantonese characters.
    tagset : str, {"universal", "hkcancor"}
        The part-of-speech tagset that the returned tags are in.
        Supported options:

        * ``"hkcancor"``, for the tagset used by the original HKCanCor data.
          There are over 100 tags, 46 of which are described at
          http://compling.hss.ntu.edu.sg/hkcancor/.
        * ``"universal"`` (default option), for the Universal Dependencies v2
          tagset. There are 17 tags; see
          https://universaldependencies.org/u/pos/index.html.
          Internally, this option applies
          :func:`~pycantonese.pos_tagging.hkcancor_to_ud` to convert HKCanCor
          tags to UD tags.

    Returns
    -------
    list[tuple[str, str]]
        The segmented sentence/phrase where each word is paired with its
        predicted POS tag.

    Raises
    ------
    TypeError
        If the input is a string (e.g., an unsegmented string of Cantonese).
    ValueError
        If the ``tagset`` argument is not one of the allowed options from
        ``{"universal", "hkcancor"}``.

    Examples
    --------
    >>> words = ['我', '噚日', '買', '嗰', '對', '鞋', '。']  # I bought that pair of shoes yesterday.
    >>> pos_tag(words)
    [('我', 'PRON'), ('噚日', 'ADV'), ('買', 'VERB'), ('嗰', 'PRON'), ('對', 'NOUN'), ('鞋', 'NOUN'), ('。', 'PUNCT')]
    >>> pos_tag(words, tagset="hkcancor")
    [('我', 'R'), ('噚日', 'T'), ('買', 'V'), ('嗰', 'R'), ('對', 'Q'), ('鞋', 'N'), ('。', '。')]
    """  # noqa: E501
    if type(words) == str:
        raise TypeError(
            f"Input must be a list of segmented words, not a string: {words}"
        )
    tags = _get_tagger().tag(words)
    if tagset == "universal":
        tags = [hkcancor_to_ud(tag) for tag in tags]
    elif tagset != "hkcancor":
        raise ValueError(f"tagset must be one of {{'universal', 'hkcancor'}}: {tagset}")
    return list(zip(words, tags))



"""Stop words for Cantonese."""

_STOP_WORDS = """
一啲
一定
不如
不過
之後
乜
乜嘢
人哋
但係
你
你哋
佢
佢哋
係
個
其他
冇
再
到
即
即係
原來
去
又
可以
可能
同
同埋
吖
呀
呢
咁
咗
咩
咪
哦
哩
哩個
哩啲
哩度
哩樣
唔
唔使
唔係
啊
啲
喎
喺
喺度
嗯
嗰
嗰個
嗰啲
嗰度
嘅
嘢
噉
噉樣
因為
多
太
好
如果
就
已經
幾
幾多
得
想
應該
成日
我
我哋
或者
所以
最
會
有
有冇
有啲
未
梗係
然之後
由
真係
睇
知
而
而家
自己
要
覺得
話
諗
講
譬如
跟住
返
過
邊個
都
點
點樣
點解
""".strip().split()


def stop_words(add=None, remove=None):
    """Return Cantonese stop words.

    .. versionadded:: 2.2.0

    Parameters
    ----------
    add : iterable[str], optional
        Stop words to add.
    remove : iterable[str], optional
        Stop words to remove.

    Returns
    -------
    set[str]

    Examples
    --------
    >>> stop_words_1 = stop_words()
    >>> len(stop_words_1)
    104
    >>> '香港' in stop_words_1
    False
    >>> stop_words_1  # doctest: +SKIP
    {'一啲', '一定', '不如', '不過', ...}
    >>>
    >>> stop_words_2 = stop_words(add='香港')
    >>> len(stop_words_2)
    105
    >>> '香港' in stop_words_2
    True
    """
    _stop_words = set(_STOP_WORDS)
    if add:
        if isinstance(add, str):
            _stop_words.add(add)
        else:
            # assume "add" is an iterable of strings
            _stop_words |= set(add)
    if remove:
        if isinstance(remove, str):
            _stop_words.remove(remove)
        else:
            # assume "remove" is an iterable of strings
            _stop_words -= set(remove)
    return _stop_words



import dataclasses
from typing import List


ONSETS = {
    "b",
    "d",
    "g",
    "gw",
    "z",
    "p",
    "t",
    "k",
    "kw",
    "c",
    "m",
    "n",
    "ng",
    "f",
    "h",
    "s",
    "l",
    "w",
    "j",
    "",
}

NUCLEI = {"aa", "a", "i", "yu", "u", "oe", "e", "eo", "o", "m", "ng"}

CODAS = {"p", "t", "k", "m", "n", "ng", "i", "u", ""}

TONES = {"1", "2", "3", "4", "5", "6"}


@dataclasses.dataclass
class Jyutping:
    """Jyutping representation of a Chinese/Cantonese character.

    Attributes
    ----------
    onset : str
        Onset
    nucleus : str
        Nucleus
    coda : str
        Coda
    tone : str
        Tone
    """

    __slots__ = ("onset", "nucleus", "coda", "tone")
    onset: str
    nucleus: str
    coda: str
    tone: str

    def __str__(self):
        """Combine onset + nucleus + coda + tone."""
        return f"{self.onset}{self.nucleus}{self.coda}{self.tone}"

    @property
    def final(self):
        """Return the final (= nucleus + coda)."""
        return f"{self.nucleus}{self.coda}"


def parse_jyutping(jp_str) -> List[Jyutping]:
    """Parse Jyutping romanization into onset, nucleus, code, and tone.

    Parameters
    ----------
    jp_str : str
        Jyutping romanization for one or multiple characters.

    Returns
    -------
    List[Jyutping]

    Raises
    ------
    ValueError
        If the Jyutping romanization is illegal (e.g., with unrecognized
        elements).

    Examples
    --------
    >>> parse_jyutping("gwong2dung1waa2")  # 廣東話, Cantonese
    [Jyutping(onset='gw', nucleus='o', coda='ng', tone='2'),
     Jyutping(onset='d', nucleus='u', coda='ng', tone='1'),
     Jyutping(onset='w', nucleus='aa', coda='', tone='2')]
    """
    if not jp_str:
        return []

    # check jp_str as a valid argument string
    if not isinstance(jp_str, str):
        raise ValueError("argument needs to be a string -- " + repr(jp_str))
    jp_str = jp_str.lower()

    # parse jp_str as multiple jp strings
    jp_list = []
    jp_current = ""
    for c in jp_str:
        jp_current = jp_current + c
        if c.isdigit():
            jp_list.append(jp_current)
            jp_current = ""

    if not jp_str[-1].isdigit():
        # TODO: error msg should be "no invalid tone detected" or something?
        raise ValueError("tone error -- " + repr(jp_str[-1]))

    jp_parsed_list = []

    for jp in jp_list:

        if len(jp) < 2:
            raise ValueError(
                "jyutping string has fewer than " "2 characters -- " + repr(jp)
            )

        tone = jp[-1]
        cvc = jp[:-1]

        # tone
        if tone not in TONES:
            raise ValueError("tone error -- " + repr(jp))

        # coda
        if not (cvc[-1] in "ieaouptkmng"):
            raise ValueError("coda error -- " + repr(jp))

        if cvc in ["m", "n", "ng", "i", "e", "aa", "o", "u"]:
            jp_parsed_list.append(Jyutping("", cvc, "", tone))
            continue
        elif cvc[-2:] == "ng":
            coda = "ng"
            cv = cvc[:-2]
        elif (
            (cvc[-1] in "ptkmn")
            or ((cvc[-1] == "i") and (cvc[-2] in "eaou"))
            or ((cvc[-1] == "u") and (cvc[-2] in "ieao"))
        ):
            coda = cvc[-1]
            cv = cvc[:-1]
        else:
            coda = ""
            cv = cvc

        # nucleus, and then onset
        nucleus = ""

        while cv[-1] in "ieaouy":
            nucleus = cv[-1] + nucleus
            cv = cv[:-1]
            if not cv:
                break

        if not nucleus:
            raise ValueError("nucleus error -- " + repr(jp))

        onset = cv

        if onset not in ONSETS:
            raise ValueError("onset error -- " + repr(jp))

        jp_parsed_list.append(Jyutping(onset, nucleus, coda, tone))

    return jp_parsed_list


def _parse_final(final):
    """Parse a final into its nucleus and coda.

    Parameters
    ----------
    final : str

    Returns
    -------
    tuple[str]
    """
    for i in range(1, len(final) + 1):
        possible_nucleus = final[:i]
        possible_coda = final[i:]

        if (possible_nucleus in NUCLEI) and (possible_coda in CODAS):
            return possible_nucleus, possible_coda
    return None



import pytest

from pycantonese.jyutping.parse_jyutping import ONSETS, TONES
from pycantonese.jyutping.tipa import ONSETS_TIPA, TONES_TIPA, jyutping_to_tipa


# TODO def test_correct_nucleus_set():
# TODO def test_correct_coda_set():


def test_correct_onset_set():
    assert set(ONSETS_TIPA.keys()) == ONSETS


def test_correct_tone_set():
    assert set(TONES_TIPA.keys()) == TONES


@pytest.mark.parametrize("input_", ["", None])
def test_null_input(input_):
    assert jyutping_to_tipa(input_) == []


def test_jyutping_to_tipa():
    assert jyutping_to_tipa("m4goi1") == ["\\s{m}21", "kOY55"]
    assert jyutping_to_tipa("gwong2dung1waa2") == [
        "k\\super w ON25",
        "tUN55",
        "wa25",
    ]



"""This script downloads the rime-cantonese data as JSON files."""

import collections
import io
import json
import logging
import os
import tempfile
import zipfile

from pycantonese.jyutping.parse_jyutping import parse_jyutping
from pycantonese.util import _split_chars_with_alphanum


try:
    import requests
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "This script requires the 'requests' package. "
        "Run `pip install requests` and then re-run this script."
    )


_BRANCH = "2020.09.09"
_URL = f"https://github.com/jacksonllee/rime-cantonese/archive/{_BRANCH}.zip"

_THIS_DIR = os.path.abspath(os.path.dirname(__file__))

_DataFile = collections.namedtuple("_DataFile", "yaml_filename, json_filename")
_CHARS_JYUTPING_FILES = (
    _DataFile("jyut6ping3.lettered.dict.yaml", "lettered.json"),
    _DataFile("jyut6ping3.dict.yaml", "chars_to_jyutping.json"),
)
_CHARS_FILES = (
    # Not yet using the "phrases" data.
    # _DataFile("jyut6ping3.phrase.dict.yaml", "phrases.json"),
    # The "maps" data has the Open Data Commons Open Database License (ODbL)
    # that has a share-alike clause -- PyCantonese can't use this data.
    # _DataFile("jyut6ping3.maps.dict.yaml", "maps.json"),
)
_RESEGMENTED_FILENAME = "resegmented.txt"


def _download_chars_jyutping_data(yaml_filename, json_filename):
    chars_to_jyutping_dicts = collections.defaultdict(dict)
    with tempfile.TemporaryDirectory() as temp_dir:
        zip_path = os.path.join(temp_dir, "file.zip")

        with requests.Session() as session, open(zip_path, "wb") as f:
            response = session.get(_URL)
            f.write(response.content)

        with zipfile.ZipFile(zip_path) as zip_file:
            data_path = os.path.join(f"rime-cantonese-{_BRANCH}", yaml_filename)
            with zip_file.open(data_path) as data_file:
                for line in io.TextIOWrapper(data_file, encoding="utf-8"):
                    line = line.strip()
                    if not line or line.startswith("#") or "\t" not in line:
                        continue

                    parts = line.split("\t")
                    if len(parts) == 2:
                        chars, jyutping = parts
                        # Default frequency is 0.07:
                        # https://github.com/CanCLID/ToJyutping/blob/be4f40d45b9780b3f0e2ffe53eff21df565ffc3f/preprocess.py#L18
                        freq = 0.07
                    elif len(parts) == 3:
                        chars, jyutping, freq_str = parts
                        # Reference for retrieving frequency:
                        # https://github.com/CanCLID/ToJyutping/blob/be4f40d45b9780b3f0e2ffe53eff21df565ffc3f/preprocess.py#L6-L16
                        try:
                            if freq_str.endswith("%"):
                                freq = float(freq_str[:-1]) * 0.01
                            else:
                                freq = float(freq_str)
                        except (ValueError, TypeError):
                            freq = 0.0
                    else:
                        continue

                    jyutping = jyutping.replace(" ", "")
                    chars_to_jyutping_dicts[chars][jyutping] = freq

    chars_to_jyutping = {}

    for chars, jyutping_dict in chars_to_jyutping_dicts.items():
        if len(jyutping_dict) == 1:
            jyutping = list(jyutping_dict.keys())[0]
        else:
            jyutping = sorted(
                jyutping_dict.items(),
                key=lambda x: x[1],  # Targeting the frequency.
                reverse=True,
            )[0][0]
        chars_to_jyutping[chars] = jyutping

    with open(os.path.join(_THIS_DIR, json_filename), "w", encoding="utf8") as f:
        json.dump(chars_to_jyutping, f, indent=4, ensure_ascii=False)


def _download_chars_data(yaml_filename, json_filename):
    chars = []
    with tempfile.TemporaryDirectory() as temp_dir:
        zip_path = os.path.join(temp_dir, "file.zip")

        with requests.Session() as session, open(zip_path, "wb") as f:
            response = session.get(_URL)
            f.write(response.content)

        with zipfile.ZipFile(zip_path) as zip_file:
            data_path = os.path.join(f"rime-cantonese-{_BRANCH}", yaml_filename)
            with zip_file.open(data_path) as data_file:
                for line in io.TextIOWrapper(data_file, encoding="utf-8"):
                    line = line.strip()
                    if (
                        not line
                        or line.startswith("#")
                        # or line[0].isalpha()
                        or line[0] in "-."
                        or ": " in line
                    ):
                        continue
                    chars.append(line)

    with open(os.path.join(_THIS_DIR, json_filename), "w", encoding="utf8") as f:
        json.dump(chars, f, indent=4, ensure_ascii=False)


def _resegment_chars_jyutping_data(json_filename):
    json_path = os.path.join(_THIS_DIR, json_filename)

    with open(json_path, encoding="utf8") as f:
        chars_to_jyutping = json.load(f)

    with open(os.path.join(_THIS_DIR, _RESEGMENTED_FILENAME), encoding="utf8") as f:
        resegmented = {}
        for line in f:
            line = line.strip()
            if not line or line.startswith("# ") or " " not in line:
                continue
            resegmented[line.replace(" ", "")] = line

    new_chars_to_jyutping = {}

    for chars, jp in chars_to_jyutping.items():
        if chars in resegmented:
            chars_split = _split_chars_with_alphanum(chars)
            jp_split = parse_jyutping(jp)

            # Don't bother if we can't match each jyutping syllable
            # with each Cantonese character.
            if len(chars_split) != len(jp_split):
                new_chars_to_jyutping[chars] = jp

            else:
                new_words = resegmented[chars].split()
                i = 0
                for new_word in new_words:

                    # If this new word already exists in the original
                    # mapping, don't re-add it to the new map, or else
                    # we risk altering this word's jyutping representation
                    # (some Cantonese words/characters have multiple
                    # pronunciations, and we've already chosen the more
                    # frequent one according to the rime-cantonese source).
                    if new_word in chars_to_jyutping:
                        i += len(_split_chars_with_alphanum(new_word))
                        continue

                    new_jp_for_word = ""
                    for _ in range(len(_split_chars_with_alphanum(new_word))):
                        new_jp_for_word += str(jp_split[i])
                        i += 1
                    new_chars_to_jyutping[new_word] = new_jp_for_word
        else:
            new_chars_to_jyutping[chars] = jp

    with open(json_path, "w", encoding="utf8") as f:
        json.dump(new_chars_to_jyutping, f, indent=4, ensure_ascii=False)


def main():
    for f in _CHARS_JYUTPING_FILES:
        logging.info("Downloading %s as %s", f.yaml_filename, f.json_filename)
        _download_chars_jyutping_data(f.yaml_filename, f.json_filename)
        logging.info("Resegmenting %s", f.json_filename)
        _resegment_chars_jyutping_data(f.json_filename)
    for f in _CHARS_FILES:
        logging.info("Downloading %s as %s", f.yaml_filename, f.json_filename)
        _download_chars_data(f.yaml_filename, f.json_filename)


if __name__ == "__main__":
    logging.basicConfig(level="INFO")
    main()



import pytest

from pycantonese.pos_tagging.hkcancor_to_ud import hkcancor_to_ud, _MAP


@pytest.mark.parametrize(
    "tag, expected",
    [
        ("V", "VERB"),
        ("v", "VERB"),
        ("V ", "VERB"),
        ("V1", "VERB"),
        ("foobar", "X"),
    ],
)
def test_hkcancor_to_ud(tag, expected):
    assert hkcancor_to_ud(tag) == expected


def test_hkcancor_to_ud_all_tags():
    assert hkcancor_to_ud() == _MAP



import pytest

from pycantonese import characters_to_jyutping


@pytest.mark.parametrize(
    "chars, expected",
    [
        (
            "香港人講廣東話。",
            [
                ("香港人", "hoeng1gong2jan4"),
                ("講", "gong2"),
                ("廣東話", "gwong2dung1waa2"),
                ("。", None),
            ],
        ),
        (
            "佢成日呃like",
            [("佢", "keoi5"), ("成日", "seng4jat6"), ("呃like", "ngaak1lai1")],
        ),
        ("蛋", [("蛋", "daan2")]),
        ("蛋糕", [("蛋糕", "daan6gou1")]),
    ],
)
def test_characters_to_jyutping(chars, expected):
    actual = characters_to_jyutping(chars)
    assert actual == expected
